import os
import tempfile

import numpy as np
import pytest
from osgeo import gdal

from tests.helpers import gdal_dataset_of_region, gdal_multiband_dataset_with_data, gdal_dataset_with_data
from yirgacheffe.window import Area, PixelScale, Window
from yirgacheffe.layers import RasterLayer, InvalidRasterBand
from yirgacheffe.rounding import round_up_pixels
from yirgacheffe.operators import DataType


# There is a lot of "del" in this file, due to a combination of gdal having no way
# to explicitly close a file beyond forcing the gdal object's deletion, and Windows
# getting upset that it tries to clear up the TemporaryDirectory and there's an open
# file handle within that directory.

def test_make_basic_layer() -> None:
    area = Area(-10, 10, 10, -10)
    dataset = gdal_dataset_of_region(area, 0.02)

    # The context manager routines are just called on the base class, so ensure
    # they do call the GDAL close method, otherwise all that is for nothing
    original_ref = dataset.Close
    close_called = []
    def mocked_close():
        close_called.append(True)
        original_ref()
    dataset.Close = mocked_close

    with RasterLayer(dataset) as layer:
        assert layer.area == area
        assert layer.pixel_scale == (0.02, -0.02)
        assert layer.geo_transform == (-10, 0.02, 0.0, 10, 0.0, -0.02)
        assert layer.window == Window(0, 0, 1000, 1000)

    assert close_called

def test_layer_from_null() -> None:
    # Seems a petty test, but gdal doesn't throw exceptions
    # so you often get None datasets if you're not careful
    with pytest.raises(ValueError):
        with RasterLayer(None) as _layer:
            pass

def test_layer_from_nonexistent_file() -> None:
    with pytest.raises(FileNotFoundError):
        _ = RasterLayer.layer_from_file("this_file_does_not_exist.tif")

def test_open_file() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        dataset = gdal_dataset_of_region(area, 0.02, filename=path)
        dataset.Close()
        assert os.path.exists(path)
        with RasterLayer.layer_from_file(path) as layer:
            assert layer.area == area
            assert layer.pixel_scale == (0.02, -0.02)
            assert layer.geo_transform == (-10, 0.02, 0.0, 10, 0.0, -0.02)
            assert layer.window == Window(0, 0, 1000, 1000)
            del layer

@pytest.mark.parametrize("initial_area",
    [
        Area(left=10.0, top=10.0, right=10.1, bottom=9.9),
        Area(left=-10.0, top=10.0, right=-9.9, bottom=9.9),
        Area(left=10.0, top=-10.0, right=10.1, bottom=-10.1),
        Area(left=-10.0, top=-10.0, right=-9.9, bottom=-10.1),
        Area(left=-0.1, top=0.1, right=0.1, bottom=-0.1),
    ]
)
def test_empty_layers_are_pixel_aligned(initial_area):
    scale = PixelScale(0.000898315284120,-0.000898315284120)

    expanded_area = initial_area.grow(0.1)

    initial_layer = RasterLayer.empty_raster_layer(initial_area, scale, gdal.GDT_Float64)

    pixel_width = (initial_layer.area.right - initial_layer.area.left) / scale.xstep
    assert round_up_pixels(pixel_width, scale.xstep) == initial_layer.window.xsize

    expanded_layer = RasterLayer.empty_raster_layer(expanded_area, scale, gdal.GDT_Float64)
    pixel_width = (expanded_layer.area.right - expanded_layer.area.left) / scale.xstep
    assert round_up_pixels(pixel_width, scale.xstep) == expanded_layer.window.xsize

def test_empty_layer_from_raster():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    empty = RasterLayer.empty_raster_layer_like(source)
    assert empty.pixel_scale == source.pixel_scale
    assert empty.projection == source.projection
    assert empty.window == source.window
    assert empty.datatype == source.datatype
    assert empty.geo_transform == source.geo_transform
    assert empty._dataset.GetRasterBand(1).GetNoDataValue() is None

@pytest.mark.parametrize("nodata", [0, 0.0, 2, 2.0])
def test_empty_layer_from_raster_with_no_data_value(nodata):
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    empty = RasterLayer.empty_raster_layer_like(source, nodata=nodata)
    assert empty.pixel_scale == source.pixel_scale
    assert empty.projection == source.projection
    assert empty.window == source.window
    assert empty.datatype == source.datatype
    assert empty.geo_transform == source.geo_transform
    assert empty._dataset.GetRasterBand(1).GetNoDataValue() == nodata

def test_empty_layer_from_raster_with_new_smaller_area():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    smaller_area = Area(-1, 1, 1, -1)
    empty = RasterLayer.empty_raster_layer_like(source, area=smaller_area)
    assert empty.pixel_scale == source.pixel_scale
    assert empty.projection == source.projection
    assert empty.window == Window(0, 0, 100, 100)
    assert empty.datatype == source.datatype
    assert empty.geo_transform == (-1.0, 0.02, 0.0, 1.0, 0.0, -0.02)

def test_empty_layer_from_raster_new_datatype():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    assert source.datatype == DataType.Byte
    empty = RasterLayer.empty_raster_layer_like(source, datatype=gdal.GDT_Float64)
    assert empty.pixel_scale == source.pixel_scale
    assert empty.projection == source.projection
    assert empty.window == source.window
    assert empty.datatype == DataType.Float64

def test_empty_layer_from_raster_with_window():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    original_window = source.window

    source.set_window_for_intersection(Area(-1, 1, 1, -1))
    assert source.window < original_window

    empty = RasterLayer.empty_raster_layer_like(source)
    assert empty.pixel_scale == source.pixel_scale
    assert empty.projection == source.projection
    assert empty.window.xoff == 0
    assert empty.window.yoff == 0
    assert empty.window.xsize == source.window.xsize
    assert empty.window.ysize == source.window.ysize

def test_layer_with_positive_offset():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.02, 20 / 0.02)

    source.offset_window_by_pixels(5)
    assert source.area == Area(-10 - (5 * 0.02), 10 + (5 * 0.02), 10 + (5 * 0.02), -10 - (5 * 0.02))
    assert source.window == Window(-5, -5, int(20 / 0.02) + 10, int(20 / 0.02) + 10)

def test_layer_with_zero_offset():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.02, 20 / 0.02)

    source.offset_window_by_pixels(0)
    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.02, 20 / 0.02)

def test_layer_with_negative_offset():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.02, 20 / 0.02)

    source.offset_window_by_pixels(-5)
    assert source.area == Area(-10 + (5 * 0.02), 10 - (5 * 0.02), 10 - (5 * 0.02), -10 + (5 * 0.02))
    assert source.window == Window(5, 5, int(20 / 0.02) - 10, int(20 / 0.02) - 10)

def test_layer_with_excessive_negative_offset():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    with pytest.raises(ValueError):
        source.offset_window_by_pixels(-9999)

def test_layer_offsets_accumulate():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.02, 20 / 0.02)

    source.offset_window_by_pixels(5)
    source.offset_window_by_pixels(-5)

    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.02, 20 / 0.02)

def test_scale_up():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.2))
    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.2, 20 / 0.2)

    new_pixel_scale = PixelScale(0.1, -0.1)
    scaled = RasterLayer.scaled_raster_from_raster(
        source,
        new_pixel_scale
    )
    assert scaled.area == Area(-10, 10, 10, -10)
    assert scaled.window == Window(0, 0, 20 / 0.1, 20 / 0.1)
    assert scaled.sum() == source.sum() * 4

def test_scale_down():
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.2))
    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.2, 20 / 0.2)

    new_pixel_scale = PixelScale(0.5, -0.5)
    scaled = RasterLayer.scaled_raster_from_raster(
        source,
        new_pixel_scale
    )
    assert scaled.area == Area(-10, 10, 10, -10)
    assert scaled.window == Window(0, 0, 20 / 0.5, 20 / 0.5)
    # because we're dropping pixels, it's not easy to do this comparison
    # but at least make sure its less
    assert scaled.sum() < source.sum()

@pytest.mark.parametrize("size,expect_success",
    [
        ((5, 5), True),
        ((5, 1), True),
        ((1, 5), True),
        ((5, 0), False),
        ((0, 5), False),
    ]
)
def test_read_array_size(size, expect_success):
    source = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.2))
    assert source.area == Area(-10, 10, 10, -10)
    assert source.window == Window(0, 0, 20 / 0.2, 20 / 0.2)

    if expect_success:
        data = source.read_array(0, 0, size[0], size[1])
        assert data.shape == (size[1], size[0])
    else:
        with pytest.raises(ValueError):
            _ = source.read_array(0, 0, size[0], size[1])

def test_invalid_band() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        dataset = gdal_dataset_of_region(area, 0.02, filename=path)
        dataset.Close()
        assert os.path.exists(path)
        with pytest.raises(InvalidRasterBand):
            _ = RasterLayer.layer_from_file(path, band=42)

def test_multiband_raster() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
    data2 = np.array([[10.0, 20.0, 30.0, 40.0], [50.0, 60.0, 70.0, 80.0]])

    datas = [data1, data2]
    dataset = gdal_multiband_dataset_with_data((0.0, 0.0), 0.02, datas)
    layer1 = RasterLayer(dataset, band=1)
    layer2 = RasterLayer(dataset, band=2)

    layers = [layer1, layer2]
    for i in range(2):
        data = datas[i]
        layer = layers[i]
        actual = layer.read_array(0, 0, 4, 2)
        assert (data == actual).all()

def test_read_array_is_numpy():
    # This test will fail if we use say the MLX backend without casting it back to numpy
    data = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
    dataset = gdal_dataset_with_data((0.0, 0.0), 0.02, data)
    with RasterLayer(dataset) as layer1:
        actual = layer1.read_array(0, 0, 4, 2).astype(int)
        expected = data.astype(int)
        assert (actual == expected).all



import os
import tempfile

import numpy as np
from osgeo import gdal

from tests.helpers import gdal_dataset_with_data
from yirgacheffe.layers import RasterLayer
from yirgacheffe.window import Area, PixelScale

def test_simple_two_band_image() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        target_path = os.path.join(tempdir, "target.tif")

        # Create an output raster layer with a number of bands
        bands = 4
        target = RasterLayer.empty_raster_layer(
            Area(-1, 1, 1, -1),
            PixelScale(1.0, -1.0),
            gdal.GDT_Byte,
            filename=target_path,
            bands=bands
        )

        # Create a set of rasters in turn to fill each band
        for i in range(bands):
            data1 = np.full((2, 2), i+1)
            layer1 = RasterLayer(gdal_dataset_with_data((-1.0, 1.0), 1.0, data1))
            layer1.save(target, band=i+1)

        # force things to disk
        target.close()

        #check they do what we expect
        for i in range(bands):
            o = RasterLayer.layer_from_file(target_path, band=i+1)
            assert o.sum() == (4 * (i + 1))

def test_stack_tifs_with_area_match() -> None:
    with tempfile.TemporaryDirectory() as tempdir:

        # create a set of tifs that have some intersection (note the
        # slight alignment offset when we create them)
        bands = 4
        source_layers = []
        for i in range(bands):
            data1 = np.full((100, 100), i+1)
            layer1 = RasterLayer(gdal_dataset_with_data((-100+i, 100+i), 1, data1))
            source_layers.append(layer1)

        intersection = RasterLayer.find_intersection(source_layers)
        for layer in source_layers:
            layer.set_window_for_intersection(intersection)

        layer = source_layers[-1]
        assert layer.window.xsize == 100 - (bands - 1)
        assert layer.window.ysize == 100 - (bands - 1)

        target_path = os.path.join(tempdir, "target.tif")
        target = RasterLayer.empty_raster_layer_like(layer, filename=target_path, bands=bands)
        for i in range(bands):
            source_layers[i].save(target, band=i+1)

        # force things to disk
        target.close()

        #check they do what we expect
        for i in range(bands):
            o = RasterLayer.layer_from_file(target_path, band=i+1)
            assert o.sum() == ((100 - (bands - 1)) * (100 - (bands - 1)) * (i + 1))



from math import ceil, floor
from typing import Any, Tuple

import h3
import numpy as np
from yirgacheffe.operators import DataType

from ..rounding import round_up_pixels
from ..window import Area, MapProjection, Window
from .base import YirgacheffeLayer
from .._backends import backend

class H3CellLayer(YirgacheffeLayer):

    def __init__(self, cell_id: str, projection: MapProjection):
        if not h3.is_valid_cell(cell_id):
            raise ValueError(f"{cell_id} is not a valid H3 cell identifier")
        self.cell_id = cell_id
        self.zoom = h3.get_resolution(cell_id)

        self.cell_boundary = h3.cell_to_boundary(cell_id)

        abs_xstep, abs_ystep = abs(projection.xstep), abs(projection.ystep)
        area = Area(
            left=(floor(min(x[1] for x in self.cell_boundary) / abs_xstep) * abs_xstep),
            top=(ceil(max(x[0] for x in self.cell_boundary) / abs_ystep) * abs_ystep),
            right=(ceil(max(x[1] for x in self.cell_boundary) / abs_xstep) * abs_xstep),
            bottom=(floor(min(x[0] for x in self.cell_boundary) / abs_ystep) * abs_ystep),
        )

        # Statistically, most hex tiles will be within a projection, but some will wrap around
        # the edge, so check if we're hit one of those cases.
        self.centre = h3.cell_to_latlng(cell_id)

        # Due to time constraints, and that most geospatial data we are working with stops before the poles
        # I'm going to explicitly not handle the poles right now, where you get different distortions from
        # at the 180 line:
        if not area.bottom < self.centre[0] < area.top:
            raise NotImplementedError("Distortion at poles not currently handled")

        # For wrap around due to hitting the longitunal wrap, we currently just do the naive thing and
        # project a band for the full width of the projection. I have a plan to fix this, as I'd like layers
        # to have sublayers, allowing us to handle vector layers more efficiently, but curently we have a
        # deadline, and given how infrequently we hit this case, doing the naive thing for now is sufficient
        if (abs(area.left - area.right)) > 180.0:
            left = (-180.0 / abs_xstep) * abs_xstep
            if left < -180.0:
                left += abs_xstep
            right = (180.0 / abs_xstep) * abs_xstep
            if right > 180.0:
                right -= abs_xstep
            area = Area(
                left=left,
                right=right,
                top=area.top,
                bottom=area.bottom,
            )

        super().__init__(area, projection)
        self._window = Window(
            xoff=0,
            yoff=0,
            xsize=round_up_pixels((area.right - area.left) / projection.xstep, abs_xstep),
            ysize=round_up_pixels((area.bottom - area.top) / projection.ystep, abs_ystep),
        )
        self._raster_xsize = self.window.xsize
        self._raster_ysize = self.window.ysize

        # see comment in read_array for why we're doing this
        sorted_lats = [x[0] for x in self.cell_boundary]
        sorted_lats.sort()
        self._raster_safe_bounds = (
            (sorted_lats[-2] / abs_ystep) * abs_ystep,
            (sorted_lats[1] / abs_ystep) * abs_ystep,
        )

    @property
    def _raster_dimensions(self) -> Tuple[int,int]:
        return (self._raster_xsize, self._raster_ysize)

    @property
    def datatype(self) -> DataType:
        return DataType.Float64

    def _read_array_with_window(
        self,
        xoffset: int,
        yoffset: int,
        xsize: int,
        ysize: int,
        window: Window,
    ) -> Any:
        assert self._projection is not None

        if (xsize <= 0) or (ysize <= 0):
            raise ValueError("Request dimensions must be positive and non-zero")

        # We have two paths: one for the common case where the hex cell doesn't cross 180˚ longitude,
        # and another case for where it does
        max_width_projection = self.area.right - self.area.left
        if max_width_projection < 180:

            target_window = Window(
                window.xoff + xoffset,
                window.yoff + yoffset,
                xsize,
                ysize
            )
            source_window = Window(
                xoff=0,
                yoff=0,
                xsize=self._raster_xsize,
                ysize=self._raster_ysize,
            )
            try:
                intersection = Window.find_intersection([source_window, target_window])
            except ValueError:
                return 0.0

            subset = np.zeros((intersection.ysize, intersection.xsize))

            start_x = self.area.left + ((intersection.xoff - self.window.xoff) * self._projection.xstep)
            start_y = self.area.top + ((intersection.yoff - self.window.yoff) * self._projection.ystep)

            for ypixel in range(intersection.ysize):
                # The latlng_to_cell is quite expensive, so ideally we want to avoid
                # calling latlng_to_cell for every pixel, though I do want to do that
                # rather than infer from the cell_boundary, as this way I'm ensured
                # I get an authoritative result and no pixel falls between the cracks.

                # Originally I just tried the old-school way of finding the left-most
                # and the right-most pixel per row, and then filling between, which in
                # theory should work fine for rasterising hexagons, but in certain places
                # the map projection distorts the edges sufficiently they become concave,
                # and that approach lead to over filling, which caused issues with cells
                # overlapping.

                # The current implementation tries to hedge its bets by looking where the
                # lowest and highest edges are (stored in _raster_safe_bounds) and then
                # only doing the fill between left and right within those bounds. I suspect
                # longer term this will need some padding, but it works for the current set
                # of known failures in test_optimisation. As we hit more issues we should
                # expand that test case before tweaking here. This is quite wasteful, as it's
                # only a tiny number of cells that have convex edges, so in future it'd be
                # interesting to see if we can infer when that is.

                lat = start_y + (ypixel * self._projection.ystep)

                if self._raster_safe_bounds[0] < lat < self._raster_safe_bounds[1]:
                    # In "safe" zone, try to be clever
                    left_most = intersection.xsize + 1
                    right_most = -1

                    for xpixel in range(intersection.xsize):
                        lng = start_x + (xpixel * self._projection.xstep)
                        this_cell = h3.latlng_to_cell(lat, lng, self.zoom)
                        if this_cell == self.cell_id:
                            left_most = xpixel
                            break

                    for xpixel in range(intersection.xsize - 1, left_most - 1, -1):
                        lng = start_x + (xpixel * self._projection.xstep)
                        this_cell = h3.latlng_to_cell(lat, lng, self.zoom)
                        if this_cell == self.cell_id:
                            right_most = xpixel
                            break

                    for xpixel in range(left_most, right_most + 1, 1):
                        subset[ypixel][xpixel] = 1.0
                else:
                    # Not in safe zone, be diligent.
                    for xpixel in range(intersection.xsize):
                        lng = start_x + (xpixel * self._projection.xstep)
                        this_cell = h3.latlng_to_cell(lat, lng, self.zoom)
                        if this_cell == self.cell_id:
                            subset[ypixel][xpixel] = 1.0

            return backend.pad(
                backend.promote(subset),
                (
                    (
                        (intersection.yoff - self.window.yoff) - yoffset,
                        (ysize - ((intersection.yoff - self.window.yoff) + intersection.ysize)) + yoffset,
                    ),
                    (
                        (intersection.xoff - self.window.xoff) - xoffset,
                        xsize - ((intersection.xoff - self.window.xoff) + intersection.xsize) + xoffset,
                    )
                ),
                'constant'
            )
        else:
            # This handles the case where the cell wraps over 180˚ longitude
            res = np.zeros((ysize, xsize))

            left = min(x[1] for x in self.cell_boundary if x[1] > 0.0)
            right = max(x[1] for x in self.cell_boundary if x[1] < 0.0) + 360.0
            max_width_projection = right - left
            max_width = ceil(max_width_projection / self._projection.xstep)

            for ypixel in range(yoffset, yoffset + ysize):
                lat = self.area.top + (ypixel * self._projection.ystep)

                for xpixel in range(xoffset, min(xoffset + xsize, max_width)):
                    lng = self.area.left + (xpixel * self._projection.xstep)
                    this_cell = h3.latlng_to_cell(lat, lng, self.zoom)
                    if this_cell == self.cell_id:
                        res[ypixel - yoffset][xpixel - xoffset] = 1.0

                for xpixel in range(xoffset + xsize - 1, xoffset + xsize - max_width, -1):
                    lng = self.area.left + (xpixel * self._projection.xstep)
                    this_cell = h3.latlng_to_cell(lat, lng, self.zoom)
                    if this_cell == self.cell_id:
                        res[ypixel - yoffset][xpixel - xoffset] = 1.0
            return backend.promote(res)



from typing import Any, Union

from ..operators import DataType
from ..window import Area, MapProjection, PixelScale, Window
from .base import YirgacheffeLayer
from .._backends import backend


class ConstantLayer(YirgacheffeLayer):
    """This is a layer that will return the identity value - can be used when an input layer is
    missing (e.g., area) without having the calculation full of branches."""
    def __init__(self, value: Union[int,float]): # pylint: disable=W0231
        area = Area.world()
        super().__init__(area, None)
        self.value = float(value)

    @property
    def datatype(self) -> DataType:
        return DataType.Float64

    def check_pixel_scale(self, _scale: PixelScale) -> bool:
        return True

    def set_window_for_intersection(self, _intersection: Area) -> None:
        pass

    def read_array(self, _x: int, _y: int, width: int, height: int) -> Any:
        return backend.full((height, width), self.value)

    def _read_array_with_window(
        self,
        _x: int,
        _y: int,
        width: int,
        height: int,
        _window: Window,
    ) -> Any:
        return backend.full((height, width), self.value)

    def _read_array_for_area(
        self,
        _target_area: Area,
        _target_projection: MapProjection,
        x: int,
        y: int,
        width: int,
        height: int,
    ) -> Any:
        return self.read_array(x, y, width, height)



import pytest

# I've no idea why pylint dislikes this particular import but accepts
# other entries in the module.
from yirgacheffe.window import Window # pylint: disable=E0401, E0611

@pytest.mark.parametrize(
    "lhs,rhs,is_greater,is_equal",
    [
        # Obvious equality
        (Window(0, 0, 100, 100),     Window(0, 0, 100, 100),     False, True),
        (Window(-10, 0, 100, 100),   Window(-10, 0, 100, 100),   False, True),
        (Window(0, -10, 100, 100),   Window(0, -10, 100, 100),   False, True),
        (Window(-10, -10, 100, 100), Window(-10, -10, 100, 100), False, True),
        (Window(10, 0, 100, 100),    Window(10, 0, 100, 100),    False, True),
        (Window(0, 10, 100, 100),    Window(0, 10, 100, 100),    False, True),
        (Window(10, 10, 100, 100),   Window(10, 10, 100, 100),   False, True),

        # Obvious inequality
        (Window(0, 0, 100, 200), Window(0, 0, 100, 100), False, False),
        (Window(0, 0, 200, 100), Window(0, 0, 100, 100), False, False),
        (Window(0, 2, 100, 100), Window(0, 0, 100, 100), False, False),
        (Window(2, 0, 100, 100), Window(0, 0, 100, 100), False, False),

        # is greater is always lhs the subset
        (Window(10, 10, 10, 10), Window(0, 0, 100, 100), True, False),
        (Window(0, 0, 10, 10),   Window(0, 0, 100, 100), True, False),
        (Window(11, 11, 8, 8),   Window(10, 10, 10, 10), True, False),
        (Window(0, 0, 10, 10), Window(-1, -1, 12, 12), True, False),

        # Here the LHS is smaller, but isn't a subset
        (Window(-5, -5, 10, 10), Window(0, 0, 100, 100), False, False),
        (Window(95, 95, 10, 10), Window(0, 0, 100, 100), False, False),
        (Window(9, 9, 5, 5),     Window(10, 10, 10, 10), False, False),
        (Window(19, 19, 5, 5),   Window(10, 10, 10, 10), False, False),

        # Smaller but with bounaries touching
        (Window(0, 0, 10, 10), Window(-2, -2, 12, 12), True, False),
        (Window(0, 0, 10, 10), Window(0, 0, 11, 11), True, False),
    ]
)
def test_window_operators(lhs: Window, rhs: Window, is_greater: bool, is_equal: bool) -> None:
    assert (lhs == rhs) == is_equal
    assert (lhs != rhs) == (not is_equal)
    assert (lhs < rhs) == is_greater
    assert (rhs > lhs) == is_greater
    assert (lhs <= rhs) == (is_equal or is_greater)
    assert (rhs >= lhs) == (is_equal or is_greater)

def test_find_intersection_empty_list() -> None:
    with pytest.raises(ValueError):
        Window.find_intersection([])

def test_find_intersection_single_item() -> None:
    window = Window(10, 10, 10, 10)
    intersection = Window.find_intersection([window])
    assert intersection == window

def test_find_intersection_same() -> None:
    windows = [
        Window(-10, 10, 10, 10),
        Window(-10, 10, 10, 10),
    ]
    intersection = Window.find_intersection(windows)
    assert intersection == windows[0]

def test_find_intersection_subset() -> None:
    windows = [
        Window(0, 0, 10, 10),
        Window(1, 1, 1, 1),
    ]
    intersection = Window.find_intersection(windows)
    assert intersection == windows[1]

def test_find_intersection_overlap() -> None:
    windows = [
        Window(0, 0, 10, 10),
        Window(5, 5, 10, 10),
    ]
    intersection = Window.find_intersection(windows)
    assert intersection == Window(5, 5, 5, 5)

def test_find_intersection_distinct() -> None:
    windows = [
        Window(0, 0, 10, 10),
        Window(20, 0, 10, 10),
    ]
    with pytest.raises(ValueError):
        _ = Window.find_intersection(windows)




import h3
import numpy as np
import pytest

from yirgacheffe import WGS_84_PROJECTION
from yirgacheffe.layers import RasterLayer, H3CellLayer
from yirgacheffe.window import Area, MapProjection
import yirgacheffe.operators as yo

class NaiveH3CellLayer(H3CellLayer):
    """h3.latlng_to_cell is quite expensive when you call it thousands of times
    so the H3CellLayer has a bunch of tricks to try reduce the work done. This is a naive
    version that checks for every cell."""

    def read_array(self, xoffset, yoffset, xsize, ysize): # pylint: disable=W0237
        assert self._projection is not None
        res = np.zeros((ysize, xsize), dtype=float)
        start_x = self.area.left + (xoffset * self._projection.xstep)
        start_y = self.area.top + (yoffset * self._projection.ystep)
        for ypixel in range(ysize):
            lat = start_y + (ypixel * self._projection.ystep)
            for xpixel in range(xsize):
                lng = start_x + (xpixel * self._projection.xstep)
                this_cell = h3.latlng_to_cell(lat, lng, self.zoom)
                if this_cell == self.cell_id:
                    res[ypixel][xpixel] = 1.0
        return res

@pytest.mark.parametrize(
    "lat,lng",
    [
        (0.0, 0.0),
        (0.0, 45.0),
        (45.0, 0.0),
        (45.0, 45.0),
        (85.0, 0.0),
        (85.0, 45.0),
        (1.0, 95.0),
    ]
)
def test_h3_vs_naive(lat: float, lng: float) -> None:
    for zoom in range(5, 9):
        cell_id = h3.latlng_to_cell(lat, lng, zoom)
        projection = MapProjection(WGS_84_PROJECTION, 0.000898315284120,-0.000898315284120)
        optimised_layer = H3CellLayer(cell_id, projection)
        naive_layer = NaiveH3CellLayer(cell_id, projection)

        optimised_cell_count = optimised_layer.sum()
        naive_cell_count = naive_layer.sum()

        assert optimised_cell_count != 0
        assert optimised_cell_count == naive_cell_count

@pytest.mark.parametrize(
    "lat,lng",
    [
        (0.0, 0.0),
        (0.0, 45.0),
        (45.0, 0.0),
        (45.0, 45.0),
        (85.0, 0.0),
        (85.0, 45.0),
        (1.0, 95.0),
    ]
)
def test_h3_vs_naive_for_union(lat: float, lng: float) -> None:
    for zoom in range(7, 9):
        cell_id = h3.latlng_to_cell(lat, lng, zoom)
        projection = MapProjection(WGS_84_PROJECTION, 0.000898315284120,-0.000898315284120)
        optimised_layer = H3CellLayer(cell_id, projection)
        naive_layer = NaiveH3CellLayer(cell_id, projection)

        before_cell_count = optimised_layer.sum()

        superset_area = Area(
            left=optimised_layer.area.left - (5 * projection.xstep),
            right=optimised_layer.area.right + (5 * projection.xstep),
            top=optimised_layer.area.top - (5 * projection.ystep),
            bottom=optimised_layer.area.bottom + (5 * projection.ystep),
        )
        optimised_layer.set_window_for_union(superset_area)
        naive_layer.set_window_for_union(superset_area)

        optimised_cell_count = optimised_layer.sum()
        naive_cell_count = naive_layer.sum()

        assert optimised_cell_count == before_cell_count
        assert optimised_cell_count == naive_cell_count

@pytest.mark.parametrize(
    "lat,lng",
    [
        (0.0, 0.0),
        (0.0, 45.0),
        (45.0, 0.0),
        (45.0, 45.0),
        (85.0, 0.0),
        (85.0, 45.0),
        (1.0, 95.0),
    ]
)
def test_h3_vs_naive_for_intersection(lat: float, lng: float) -> None:
    for zoom in range(7, 9):
        cell_id = h3.latlng_to_cell(lat, lng, zoom)
        projection = MapProjection(WGS_84_PROJECTION, 0.000898315284120,-0.000898315284120)
        optimised_layer = H3CellLayer(cell_id, projection)
        naive_layer = NaiveH3CellLayer(cell_id, projection)

        before_cell_count = optimised_layer.sum()

        subset_area = Area(
            left=optimised_layer.area.left + (2 * projection.xstep),
            right=optimised_layer.area.right - (2 * projection.xstep),
            top=optimised_layer.area.top + (2 * projection.ystep),
            bottom=optimised_layer.area.bottom - (2 * projection.ystep),
        )
        optimised_layer.set_window_for_intersection(subset_area)
        naive_layer.set_window_for_intersection(subset_area)

        optimised_cell_count = optimised_layer.sum()
        naive_cell_count = naive_layer.sum()

        assert optimised_cell_count < before_cell_count
        assert optimised_cell_count == naive_cell_count

# This test exits because certain hexagons have convex edges due to projection
# distortion. This means that my original naive attempt to speed up rendering by
# just finding the left and right most pixels on a line and filling between them
# leads to over-filling some tiles, which then overlap their neighbours.
@pytest.mark.parametrize(
    "cell_id",
    [
        "87a968290ffffff",
        "87a9680e0ffffff",
        "87a96811cffffff",
        "87a968293ffffff",
        "87a968768ffffff",
        "87a968893ffffff",
        "87a968000ffffff",
        "87a96b934ffffff",
        "87a968130ffffff",
    ]
)
def test_cells_dont_overlap(cell_id):

    cluster = h3.grid_disk(cell_id, 1)
    projection = MapProjection(WGS_84_PROJECTION, 0.000898315284120,-0.000898315284120)
    layers = [H3CellLayer(x, projection) for x in cluster]

    union = RasterLayer.find_union(layers)
    for layer in layers:
        layer.set_window_for_union(union)

    calc = layers[0]
    for layer in layers[1:]:
        calc += layer
    calc = yo.where(calc > 1, 1, 0)
    assert calc.sum() == 0

    # If we didn't get an exception, then there's no over drawing



import pytest

from yirgacheffe.window import MapProjection
from yirgacheffe.rounding import MINIMAL_DEGREE_OF_INTEREST

def test_scale_from_projection() -> None:
    projection = MapProjection("PROJ", 0.1, -0.1)
    assert projection.name == "PROJ"
    assert projection.xstep == 0.1
    assert projection.ystep == -0.1

    scale = projection.scale
    assert scale.xstep == 0.1
    assert scale.ystep == -0.1

@pytest.mark.parametrize(
    "lhs,rhs,is_equal",
    [
        (MapProjection("A", 0.1, -0.1), MapProjection("A", 0.1, -0.1), True),
        (MapProjection("A", 0.1, -0.1), MapProjection("B", 0.1, -0.1), False),
        (MapProjection("A", 0.1, -0.1), MapProjection("A", 0.1, 0.1), False),
        (MapProjection("A", 0.1, -0.1), MapProjection("A", -0.1, 0.1), False),
        (MapProjection("A", 0.1, -0.1), MapProjection("A", 0.1 + (MINIMAL_DEGREE_OF_INTEREST / 2), -0.1), True),
        (MapProjection("A", 0.1, -0.1), MapProjection("A", 0.1 - (MINIMAL_DEGREE_OF_INTEREST / 2), -0.1), True),
        (MapProjection("A", 0.1, -0.1), MapProjection("A", 0.1, -0.1 + (MINIMAL_DEGREE_OF_INTEREST / 2)), True),
        (MapProjection("A", 0.1, -0.1), MapProjection("A", 0.1, -0.1 - (MINIMAL_DEGREE_OF_INTEREST / 2)), True),
    ]
)
def test_projection_equality(lhs: MapProjection, rhs : MapProjection, is_equal: bool) -> None:
    assert MINIMAL_DEGREE_OF_INTEREST > 0.0
    assert (lhs == rhs) == is_equal
    assert (lhs != rhs) == (not is_equal)



import os
import tempfile
from pathlib import Path

import numpy as np
import pytest

import yirgacheffe as yg
from yirgacheffe import WGS_84_PROJECTION
from yirgacheffe.layers import InvalidRasterBand
from yirgacheffe.window import Area, MapProjection, Window
from tests.helpers import gdal_dataset_of_region, gdal_multiband_dataset_with_data, \
    make_vectors_with_id, make_vectors_with_mutlile_ids

def test_raster_from_nonexistent_file() -> None:
    with pytest.raises(FileNotFoundError):
        _ = yg.read_raster("this_file_does_not_exist.tif")

def test_open_raster_file() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        dataset = gdal_dataset_of_region(area, 0.02, filename=path)
        dataset.Close()
        assert os.path.exists(path)

        with yg.read_raster(path) as layer:
            assert layer.area == area
            assert layer.pixel_scale == (0.02, -0.02)
            assert layer.geo_transform == (-10, 0.02, 0.0, 10, 0.0, -0.02)
            assert layer.window == Window(0, 0, 1000, 1000)

def test_open_raster_file_as_path() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = Path(tempdir) / "test.tif"
        area = Area(-10, 10, 10, -10)
        dataset = gdal_dataset_of_region(area, 0.02, filename=path)
        dataset.Close()
        assert path.exists

        with yg.read_raster(path) as layer:
            assert layer.area == area
            assert layer.pixel_scale == (0.02, -0.02)
            assert layer.geo_transform == (-10, 0.02, 0.0, 10, 0.0, -0.02)
            assert layer.window == Window(0, 0, 1000, 1000)

def test_open_multiband_raster_wrong_band() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = Path(tempdir) / "test.tif"
        data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
        data2 = np.array([[10.0, 20.0, 30.0, 40.0], [50.0, 60.0, 70.0, 80.0]])

        datas = [data1, data2]
        dataset = gdal_multiband_dataset_with_data((0.0, 0.0), 0.02, datas, filename=path)
        dataset.Close()

        with pytest.raises(InvalidRasterBand):
            _ = yg.read_raster(path, 3)

def test_open_multiband_raster() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = Path(tempdir) / "test.tif"
        data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
        data2 = np.array([[10.0, 20.0, 30.0, 40.0], [50.0, 60.0, 70.0, 80.0]])

        datas = [data1, data2]
        dataset = gdal_multiband_dataset_with_data((0.0, 0.0), 0.02, datas, filename=path)
        dataset.Close()

        for i in range(2):
            with yg.read_raster(path, i + 1) as layer:
                data = datas[i]
                actual = layer.read_array(0, 0, 4, 2)
                assert (data == actual).all()

def test_shape_from_nonexistent_file() -> None:
    with pytest.raises(FileNotFoundError):
        _ = yg.read_shape("this_file_does_not_exist.gpkg", (WGS_84_PROJECTION, (1.0, -1.0)))

def test_open_gpkg() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.gpkg")
        area = Area(-10.0, 10.0, 10.0, 0.0)
        make_vectors_with_id(42, {area}, path)

        with yg.read_shape(path, (WGS_84_PROJECTION, (1.0, -1.0))) as layer:
            assert layer.area == area
            assert layer.geo_transform == (area.left, 1.0, 0.0, area.top, 0.0, -1.0)
            assert layer.window == Window(0, 0, 20, 10)
            assert layer.map_projection == MapProjection(WGS_84_PROJECTION, 1.0, -1.0)
            assert layer.projection == WGS_84_PROJECTION

def test_open_gpkg_with_mapprojection() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.gpkg")
        area = Area(-10.0, 10.0, 10.0, 0.0)
        make_vectors_with_id(42, {area}, path)

        with yg.read_shape(path, MapProjection(WGS_84_PROJECTION, 1.0, -1.0)) as layer:
            assert layer.area == area
            assert layer.geo_transform == (area.left, 1.0, 0.0, area.top, 0.0, -1.0)
            assert layer.window == Window(0, 0, 20, 10)
            assert layer.map_projection == MapProjection(WGS_84_PROJECTION, 1.0, -1.0)
            assert layer.projection == WGS_84_PROJECTION

def test_open_gpkg_with_no_projection() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.gpkg")
        area = Area(-10.0, 10.0, 10.0, 0.0)
        make_vectors_with_id(42, {area}, path)

        with yg.read_shape(path) as layer:
            assert layer.area == area
            assert layer.projection is None
            with pytest.raises(AttributeError):
                _ = layer.geo_transform
            with pytest.raises(AttributeError):
                _ = layer.window

def test_open_gpkg_direct_scale() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = Path(tempdir) / "test.gpkg"
        area = Area(-10.0, 10.0, 10.0, 0.0)
        make_vectors_with_id(42, {area}, path)

        with yg.read_shape(path, (WGS_84_PROJECTION, (1.0, -1.0))) as layer:
            assert layer.area == area
            assert layer.geo_transform == (area.left, 1.0, 0.0, area.top, 0.0, -1.0)
            assert layer.window == Window(0, 0, 20, 10)
            assert layer.projection == WGS_84_PROJECTION

def test_open_gpkg_with_filter() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = Path(tempdir) / "test.gpkg"
        areas = {
            (Area(-10.0, 10.0, 0.0, 0.0), 42),
            (Area(0.0, 0.0, 10, -10), 43)
        }
        make_vectors_with_mutlile_ids(areas, path)

        with yg.read_shape(path, (WGS_84_PROJECTION, (1.0, -1.0)), "id_no=42") as layer:
            assert layer.area == Area(-10.0, 10.0, 0.0, 0.0)
            assert layer.geo_transform == (-10.0, 1.0, 0.0, 10.0, 0.0, -1.0)
            assert layer.window == Window(0, 0, 10, 10)

            # Because we picked one later, all pixels should be burned
            total = layer.sum()
            assert total == (layer.window.xsize * layer.window.ysize)

def test_open_shape_like() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = Path(tempdir) / "test.tif"
        area = Area(-10, 10, 10, -10)
        dataset = gdal_dataset_of_region(area, 1.0, filename=path)
        dataset.Close()
        assert os.path.exists(path)

        with yg.read_raster(path) as raster_layer:
            path = os.path.join(tempdir, "test.gpkg")
            area = Area(-10.0, 10.0, 10.0, 0.0)
            make_vectors_with_id(42, {area}, path)

            with yg.read_shape_like(path, raster_layer) as layer:
                assert layer.area == area
                assert layer.geo_transform == (area.left, 1.0, 0.0, area.top, 0.0, -1.0)
                assert layer.window == Window(0, 0, 20, 10)
                assert layer.projection == raster_layer.projection

@pytest.mark.parametrize("tiled", [False, True])
def test_empty_rasters_list(tiled):
    with pytest.raises(ValueError):
        _ = yg.read_rasters([], tiled=tiled)

@pytest.mark.parametrize("tiled", [False, True])
def test_open_two_raster_areas_side_by_side(tiled):
    with tempfile.TemporaryDirectory() as tempdir:
        path1 = Path(tempdir) / "test1.tif"
        area1 = Area(-10, 10, 10, -10)
        dataset1 = gdal_dataset_of_region(area1, 0.2, filename=path1)
        dataset1.Close()

        path2 = Path(tempdir) / "test2.tif"
        area2 = Area(10, 10, 30, -10)
        dataset2 = gdal_dataset_of_region(area2, 0.2, filename=path2)
        dataset2.Close()

        with yg.read_rasters([path1, path2], tiled=tiled) as group:
            assert group.area == Area(-10, 10, 30, -10)
            assert group.window == Window(0, 0, 200, 100)

            with yg.read_raster(path1) as raster1:
                with yg.read_raster(path2) as raster2:
                    assert group.sum() == raster1.sum() + raster2.sum()

@pytest.mark.parametrize("tiled", [False, True])
def test_open_two_raster_by_glob(tiled):
    with tempfile.TemporaryDirectory() as tempdir:
        temppath = Path(tempdir)
        path1 = temppath / "test1.tif"
        area1 = Area(-10, 10, 10, -10)
        dataset1 = gdal_dataset_of_region(area1, 0.2, filename=path1)
        dataset1.Close()

        path2 = temppath / "test2.tif"
        area2 = Area(10, 10, 30, -10)
        dataset2 = gdal_dataset_of_region(area2, 0.2, filename=path2)
        dataset2.Close()

        with yg.read_rasters(temppath.glob("*.tif"), tiled=tiled) as group:
            assert group.area == Area(-10, 10, 30, -10)
            assert group.window == Window(0, 0, 200, 100)

            with yg.read_raster(path1) as raster1:
                with yg.read_raster(path2) as raster2:
                    assert group.sum() == raster1.sum() + raster2.sum()



import numpy as np

from yirgacheffe.layers.rasters import RasterLayer
from yirgacheffe.layers.group import GroupLayer

from tests.helpers import gdal_dataset_with_data

def test_raster_without_nodata_value() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 5.0, 8.0]])
    dataset = gdal_dataset_with_data((0.0, 0.0), 0.02, data1)
    with RasterLayer(dataset) as layer:
        assert layer.nodata is None
        actual = layer.read_array(0, 0, 4, 2)
        assert np.array_equal(data1, actual, equal_nan=True)

def test_raster_with_nodata_value() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 5.0, 8.0]])
    dataset = gdal_dataset_with_data((0.0, 0.0), 0.02, data1)
    dataset.GetRasterBand(1).SetNoDataValue(5.0)
    with RasterLayer(dataset) as layer:
        assert layer.nodata == 5.0
        data1[data1 == 5.0] = np.nan
        actual = layer.read_array(0, 0, 4, 2)
        assert np.array_equal(data1, actual, equal_nan=True)

def test_raster_with_nodata_value_ignored() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 5.0, 8.0]])
    dataset = gdal_dataset_with_data((0.0, 0.0), 0.02, data1)
    dataset.GetRasterBand(1).SetNoDataValue(5.0)
    with RasterLayer(dataset, ignore_nodata=True) as layer:
        assert layer.nodata == 5.0
        actual = layer.read_array(0, 0, 4, 2)
        assert np.array_equal(data1, actual, equal_nan=True)

def test_group_layer_with_nodata_values() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 5.0, 5.0, 5.0]])
    dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1)
    dataset1.GetRasterBand(1).SetNoDataValue(5.0)

    data2 = np.array([[1.0, 1.0, 1.0, 1.0], [5.0, 6.0, 7.0, 8.0]])
    dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2)
    dataset2.GetRasterBand(1).SetNoDataValue(1.0)

    with RasterLayer(dataset1) as layer1:
        with RasterLayer(dataset2) as layer2:
            with GroupLayer([layer1, layer2]) as group:
                actual = group.read_array(0, 0, 4, 2)
                expected = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
                assert np.array_equal(expected, actual, equal_nan=True)

def test_group_layer_with_nodata_values_ignore_nodata() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 5.0, 5.0, 5.0]])
    dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1)
    dataset1.GetRasterBand(1).SetNoDataValue(5.0)

    data2 = np.array([[1.0, 1.0, 1.0, 1.0], [5.0, 6.0, 7.0, 8.0]])
    dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2)
    dataset2.GetRasterBand(1).SetNoDataValue(1.0)

    with RasterLayer(dataset1, ignore_nodata=True) as layer1:
        with RasterLayer(dataset2, ignore_nodata=True) as layer2:
            with GroupLayer([layer1, layer2]) as group:
                actual = group.read_array(0, 0, 4, 2)
                assert np.array_equal(data1, actual, equal_nan=True)



from typing import Callable, Dict

import numpy as np
import torch

from .enumeration import operators as op
from .enumeration import dtype

array_t = np.ndarray
float_t = np.float64

promote = lambda a: a
demote_array = lambda a: a
demote_scalar = lambda a: a
eval_op = lambda a: a

add_op = np.ndarray.__add__
sub_op = np.ndarray.__sub__
mul_op = np.ndarray.__mul__
truediv_op = np.ndarray.__truediv__
pow_op = np.ndarray.__pow__
eq_op = np.ndarray.__eq__
ne_op = np.ndarray.__ne__
lt_op = np.ndarray.__lt__
le_op = np.ndarray.__le__
gt_op = np.ndarray.__gt__
ge_op = np.ndarray.__ge__
and_op = np.ndarray.__and__
or_op = np.ndarray.__or__
nan_to_num = np.nan_to_num
isin = np.isin
log = np.log
log2 = np.log2
log10 = np.log10
exp = np.exp
exp2 = np.exp2
clip = np.clip
where = np.where
min_op = np.min
max_op = np.max
maximum = np.maximum
minimum = np.minimum
zeros = np.zeros
pad = np.pad
sum_op = lambda a: np.sum(a.astype(np.float64))
isscalar = np.isscalar
full = np.full
allclose = np.allclose
remainder_op = np.ndarray.__mod__
floordiv_op = np.ndarray.__floordiv__
abs_op = np.abs
floor_op = np.floor
round_op = np.round
ceil_op = np.ceil

def conv2d_op(data, weights):
    # torch wants to process dimensions of channels of width of height
    # Which is why both the data and weights get nested into two arrays here,
    # and then we have to unpack it from that nesting.

    preped_weights = np.array([[weights]])
    conv = torch.nn.Conv2d(1, 1, weights.shape, bias=False)
    conv.weight = torch.nn.Parameter(torch.from_numpy(preped_weights))
    preped_data = torch.from_numpy(np.array([[data]]))

    res = conv(preped_data)
    return res.detach().numpy()[0][0]

def dtype_to_backed(dt):
    match dt:
        case dtype.Float32:
            return np.float32
        case dtype.Float64:
            return np.float64
        case dtype.Byte:
            return np.uint8
        case dtype.Int8:
            return np.int8
        case dtype.Int16:
            return np.int16
        case dtype.Int32:
            return np.int32
        case dtype.Int64:
            return np.int64
        case dtype.UInt8:
            return np.uint8
        case dtype.UInt16:
            return np.uint16
        case dtype.UInt32:
            return np.uint32
        case dtype.UInt64:
            return np.uint64
        case _:
            raise ValueError

def backend_to_dtype(val):
    match val:
        case np.float32:
            return dtype.Float32
        case np.float64:
            return dtype.Float64
        case np.int8:
            return dtype.Int8
        case np.int16:
            return dtype.Int16
        case np.int32:
            return dtype.Int32
        case np.int64:
            return dtype.Int64
        case np.uint8:
            return dtype.Byte
        case np.uint16:
            return dtype.UInt16
        case np.uint32:
            return dtype.UInt32
        case np.uint64:
            return dtype.UInt64
        case _:
            raise ValueError

def astype_op(data, datatype):
    return data.astype(dtype_to_backed(datatype))

operator_map : Dict[op,Callable] = {
    op.ADD: np.ndarray.__add__,
    op.SUB: np.ndarray.__sub__,
    op.MUL: np.ndarray.__mul__,
    op.TRUEDIV: np.ndarray.__truediv__,
    op.POW: np.ndarray.__pow__,
    op.EQ: np.ndarray.__eq__,
    op.NE: np.ndarray.__ne__,
    op.LT: np.ndarray.__lt__,
    op.LE: np.ndarray.__le__,
    op.GT: np.ndarray.__gt__,
    op.GE: np.ndarray.__ge__,
    op.AND: np.ndarray.__and__,
    op.OR: np.ndarray.__or__,
    op.LOG: np.log,
    op.LOG2: np.log2,
    op.LOG10: np.log10,
    op.EXP: np.exp,
    op.EXP2: np.exp2,
    op.CLIP: np.clip,
    op.WHERE: np.where,
    op.MIN: np.min,
    op.MAX: np.max,
    op.MINIMUM: np.minimum,
    op.MAXIMUM: np.maximum,
    op.NAN_TO_NUM: np.nan_to_num,
    op.ISIN: np.isin,
    op.REMAINDER: np.ndarray.__mod__,
    op.FLOORDIV: np.ndarray.__floordiv__,
    op.CONV2D: conv2d_op,
    op.ABS: np.abs,
    op.ASTYPE: astype_op,
    op.FLOOR: np.floor,
    op.ROUND: np.round,
    op.CEIL: np.ceil,
}




import math
import os
import pickle
import tempfile

import numpy as np
import pytest

from tests.helpers import gdal_dataset_of_region, make_vectors_with_id
from yirgacheffe.window import Area, MapProjection, PixelScale, Window
from yirgacheffe.layers import ConstantLayer, GroupLayer, RasterLayer, RescaledRasterLayer, \
    UniformAreaLayer, VectorLayer
from yirgacheffe import WGS_84_PROJECTION


def test_pickle_raster_layer() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        layer = RasterLayer(gdal_dataset_of_region(area, 0.02, filename=path))

        p = pickle.dumps(layer)
        restore = pickle.loads(p)

        assert restore.area == area
        assert restore.pixel_scale == (0.02, -0.02)
        assert restore.geo_transform == (-10, 0.02, 0.0, 10, 0.0, -0.02)
        assert restore.window == Window(0, 0, 1000, 1000)

def test_pickle_raster_mem_layer_fails() -> None:
    area = Area(-10, 10, 10, -10)
    layer = RasterLayer(gdal_dataset_of_region(area, 0.02))

    with pytest.raises(ValueError):
        _ = pickle.dumps(layer)

def test_pickle_dyanamic_vector_layer() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.gpkg")
        area = Area(-10.0, 10.0, 10.0, 0.0)
        make_vectors_with_id(42, {area}, path)

        layer = VectorLayer.layer_from_file(path, "id_no = 42", PixelScale(1.0, -1.0), WGS_84_PROJECTION)

        p = pickle.dumps(layer)
        restore = pickle.loads(p)

        assert restore.area == area
        assert restore.geo_transform == (area.left, 1.0, 0.0, area.top, 0.0, -1.0)
        assert restore.window == Window(0, 0, 20, 10)
        assert restore.projection == WGS_84_PROJECTION

        del layer

def test_pickle_uniform_area_layer() -> None:
    pixel_scale = 0.2
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(
            math.floor(-180 / pixel_scale) * pixel_scale,
            math.ceil(90 / pixel_scale) * pixel_scale,
            (math.floor(-180 / pixel_scale) * pixel_scale) + pixel_scale,
            math.floor(-90 / pixel_scale) * pixel_scale
        )
        dataset = gdal_dataset_of_region(area, pixel_scale, filename=path)
        assert dataset.RasterXSize == 1
        assert dataset.RasterYSize == math.ceil(180 / pixel_scale)
        dataset.Close()

        layer = UniformAreaLayer.layer_from_file(path)

        p = pickle.dumps(layer)
        restore = pickle.loads(p)

        assert restore.pixel_scale == (pixel_scale, -pixel_scale)
        assert restore.area == Area(
            math.floor(-180 / pixel_scale) * pixel_scale,
            math.ceil(90 / pixel_scale) * pixel_scale,
            math.ceil(180 / pixel_scale) * pixel_scale,
            math.floor(-90 / pixel_scale) * pixel_scale
        )
        assert restore.window == Window(
            0,
            0,
            math.ceil((restore.area.right - restore.area.left) / pixel_scale),
            math.ceil((restore.area.top - restore.area.bottom) / pixel_scale)
        )

def test_pickle_group_layer() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        dataset = gdal_dataset_of_region(area, 0.2, filename=path)
        dataset.Close()

        group = GroupLayer.layer_from_directory(tempdir)
        expected = group.read_array(0, 0, 100, 100)
        assert expected.sum() != 0 # just check there is meaningful data

        p = pickle.dumps(group)
        restore = pickle.loads(p)

        assert restore.area == area
        assert restore.window == Window(0, 0, 100, 100)

        result = restore.read_array(0, 0, 100, 100)
        assert (expected == result).all()

@pytest.mark.parametrize("c", [
    (float(2.5)),
    (int(2)),
    (np.uint16(2)),
    (np.float32(2.5)),
])
def test_pickle_constant_layer(c) -> None:
    layer = ConstantLayer(c)

    p = pickle.dumps(layer)
    restore = pickle.loads(p)

    result = restore.read_array(0, 0, 1, 1)
    assert (result == np.array([[c]])).all()

@pytest.mark.parametrize("c", [
    (float(2.5)),
    (int(2)),
    (np.uint16(2)),
    (np.float32(2.5)),
])
def test_pickle_simple_calc(c) -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        layer = RasterLayer(gdal_dataset_of_region(area, 0.2, filename=path))

        calc = layer * c
        assert calc.sum() != 0
        assert calc.sum() == layer.sum() * c

        p = pickle.dumps(calc)
        restore = pickle.loads(p)
        assert calc.sum() == restore.sum()

def test_pickle_lambda_calc() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        layer = RasterLayer(gdal_dataset_of_region(area, 0.2, filename=path))

        calc = layer.numpy_apply(lambda x: x * 2.0)
        assert calc.sum() != 0
        assert calc.sum() == layer.sum() * 2

        p = pickle.dumps(calc)
        restore = pickle.loads(p)

        assert calc.sum() == restore.sum()

def test_pickle_func_calc() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        layer = RasterLayer(gdal_dataset_of_region(area, 0.2, filename=path))

        def mulex(x):
            return x * 2.0

        calc = layer.numpy_apply(mulex)
        assert calc.sum() != 0
        assert calc.sum() == layer.sum() * 2

        p = pickle.dumps(calc)
        restore = pickle.loads(p)

        assert calc.sum() == restore.sum()

def test_pickle_rescaled_raster_layer() -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.tif")
        area = Area(-10, 10, 10, -10)
        raster = RasterLayer(gdal_dataset_of_region(area, 0.02, filename=path))
        layer = RescaledRasterLayer(raster, MapProjection(WGS_84_PROJECTION, 0.01, -0.01))

        p = pickle.dumps(layer)
        restore = pickle.loads(p)

        assert restore.area == area
        assert restore.pixel_scale == (0.01, -0.01)
        assert restore.geo_transform == (-10, 0.01, 0.0, 10, 0.0, -0.01)
        assert restore.window == Window(0, 0, 2000, 2000)

        expected = restore.read_array(0, 0, 100, 100)
        assert expected.sum() != 0 # just check there is meaningful data



from __future__ import annotations
import math
from pathlib import Path
from typing import Any, Optional, Tuple, Union

import numpy as np
from osgeo import gdal

from ..constants import WGS_84_PROJECTION
from ..window import Area, MapProjection, PixelScale, Window
from ..rounding import round_up_pixels
from .base import YirgacheffeLayer
from ..operators import DataType
from .._backends import backend

class InvalidRasterBand(Exception):
    def __init__ (self, band):
        self.band = band

class RasterLayer(YirgacheffeLayer):
    """Layer provides a wrapper around a gdal dataset/band that also records offset state so that
    we can work with maps over different geographic regions but work withing a particular frame
    of reference."""

    @staticmethod
    def empty_raster_layer(
        area: Area,
        scale: PixelScale,
        datatype: Union[int, DataType],
        filename: Optional[str]=None,
        projection: str=WGS_84_PROJECTION,
        name: Optional[str]=None,
        compress: bool=True,
        nodata: Optional[Union[float,int]]=None,
        nbits: Optional[int]=None,
        threads: Optional[int]=None,
        bands: int=1
    ) -> RasterLayer:
        abs_xstep, abs_ystep = abs(scale.xstep), abs(scale.ystep)

        # We treat the provided area as aspirational, and we need to align it to pixel boundaries
        pixel_friendly_area = Area(
            left=math.floor(area.left / abs_xstep) * abs_xstep,
            right=math.ceil(area.right / abs_xstep) * abs_xstep,
            top=math.ceil(area.top / abs_ystep) * abs_ystep,
            bottom=math.floor(area.bottom / abs_ystep) * abs_ystep,
        )

        # This used to the the GDAL type, so we support that for legacy reasons
        if isinstance(datatype, int):
            datatype_arg = DataType.of_gdal(datatype)
        else:
            datatype_arg = datatype

        options = []
        if threads is not None:
            options.append(f"NUM_THREADS={threads}")
        if nbits is not None:
            options.append(f"NBITS={nbits}")

        if filename:
            driver = gdal.GetDriverByName('GTiff')
            options.append('BIGTIFF=YES')
            if compress:
                options.append('COMPRESS=LZW')
            else:
                options.append('COMPRESS=NONE')
        else:
            driver = gdal.GetDriverByName('mem')
            filename = 'mem'
            compress = False
        dataset = driver.Create(
            filename,
            round_up_pixels((pixel_friendly_area.right - pixel_friendly_area.left) / abs_xstep, abs_xstep),
            round_up_pixels((pixel_friendly_area.top - pixel_friendly_area.bottom) / abs_ystep, abs_ystep),
            bands,
            datatype_arg.to_gdal(),
            options
        )
        dataset.SetGeoTransform([
            pixel_friendly_area.left, scale.xstep, 0.0, pixel_friendly_area.top, 0.0, scale.ystep
        ])
        dataset.SetProjection(projection)
        if nodata is not None:
            dataset.GetRasterBand(1).SetNoDataValue(nodata)
        return RasterLayer(dataset, name=name)

    @staticmethod
    def empty_raster_layer_like(
        layer: Any,
        filename: Optional[Union[Path,str]]=None,
        area: Optional[Area]=None,
        datatype: Optional[Union[int, DataType]]=None,
        compress: bool=True,
        nodata: Optional[Union[float,int]]=None,
        nbits: Optional[int]=None,
        threads: Optional[int]=None,
        bands: int=1
    ) -> RasterLayer:
        width = layer.window.xsize
        height = layer.window.ysize
        if area is None:
            area = layer.area
        assert area is not None

        if datatype is not None:
            if isinstance(datatype, int):
                datatype = DataType.of_gdal(datatype)

        projection = layer.map_projection
        if projection is None:
            raise ValueError("Can not work out area without explicit pixel scale")
        abs_xstep, abs_ystep = abs(projection.xstep), abs(projection.ystep)
        width = round_up_pixels((area.right - area.left) / abs_xstep, abs_xstep)
        height = round_up_pixels((area.top - area.bottom) / abs_ystep, abs_ystep)
        geo_transform = (
            area.left, projection.xstep, 0.0, area.top, 0.0, projection.ystep
        )

        if datatype is None:
            datatype_arg = layer.datatype
        elif isinstance(datatype, int):
            datatype_arg = DataType.of_gdal(datatype)
        else:
            datatype_arg = datatype

        options = []
        if threads is not None:
            options.append(f"NUM_THREADS={threads}")
        if nbits is not None:
            options.append(f"NBITS={nbits}")

        if filename:
            driver = gdal.GetDriverByName('GTiff')
            options.append('BIGTIFF=YES')
            if compress:
                options.append('COMPRESS=LZW')
            else:
                options.append('COMPRESS=NONE')
        else:
            driver = gdal.GetDriverByName('mem')
            filename = 'mem'
            compress = False
        dataset = driver.Create(
            filename,
            width,
            height,
            bands,
            (datatype_arg).to_gdal(),
            options,
        )
        dataset.SetGeoTransform(geo_transform)
        dataset.SetProjection(projection.name)
        if nodata is not None:
            dataset.GetRasterBand(1).SetNoDataValue(nodata)

        return RasterLayer(dataset)

    @classmethod
    def scaled_raster_from_raster(
        cls,
        source: RasterLayer,
        new_pixel_scale: PixelScale,
        filename: Optional[str]=None,
        compress: bool=True,
        algorithm: int=gdal.GRA_NearestNeighbour,
    ) -> RasterLayer:
        source_dataset = source._dataset
        assert source_dataset is not None

        old_projection = source.map_projection
        assert old_projection is not None

        x_scale = old_projection.xstep / new_pixel_scale.xstep
        y_scale = old_projection.ystep / new_pixel_scale.ystep
        new_width = round_up_pixels(source_dataset.RasterXSize * x_scale,
            abs(new_pixel_scale.xstep))
        new_height = round_up_pixels(source_dataset.RasterYSize * y_scale,
            abs(new_pixel_scale.ystep))

        # in yirgacheffe we like to have things aligned to the pixel_scale, so work
        # out new top left corner
        new_left = math.floor((source.area.left / new_pixel_scale.xstep)) * new_pixel_scale.xstep
        new_top = math.ceil((source.area.top / new_pixel_scale.ystep)) * new_pixel_scale.ystep

        # now build a target dataset
        options = []
        if filename:
            driver = gdal.GetDriverByName('GTiff')
            options.append('BIGTIFF=YES')
            if compress:
                options.append('COMPRESS=LZW')
        else:
            driver = gdal.GetDriverByName('mem')
            filename = 'mem'
            compress = False
        dataset = driver.Create(
            filename,
            new_width,
            new_height,
            1,
            source.datatype.to_gdal(),
            options
        )
        dataset.SetGeoTransform((
            new_left, new_pixel_scale.xstep, 0.0,
            new_top, 0.0, new_pixel_scale.ystep
        ))
        dataset.SetProjection(source_dataset.GetProjection())

        # now use gdal to do the reprojection
        gdal.ReprojectImage(source_dataset, dataset, eResampleAlg=algorithm)

        return RasterLayer(dataset)

    @classmethod
    def layer_from_file(
        cls,
        filename: Union[Path,str],
        band: int = 1,
        ignore_nodata: bool = False,
    ) -> RasterLayer:
        try:
            dataset = gdal.Open(filename, gdal.GA_ReadOnly)
        except RuntimeError as exc:
            # With exceptions on GDAL now returns the wrong (IMHO) exception
            raise FileNotFoundError(filename) from exc
        try:
            _ = dataset.GetRasterBand(band)
        except RuntimeError as exc:
            raise InvalidRasterBand(band) from exc
        return cls(dataset, str(filename), band, ignore_nodata)

    def __init__(
        self,
        dataset: gdal.Dataset,
        name: Optional[str] = None,
        band: int = 1,
        ignore_nodata: bool = False,
    ) -> None:
        if not dataset:
            raise ValueError("None is not a valid dataset")

        transform = dataset.GetGeoTransform()
        projection = MapProjection(dataset.GetProjection(), transform[1], transform[5])
        area = Area(
            left=transform[0],
            top=transform[3],
            right=transform[0] + (dataset.RasterXSize * projection.xstep),
            bottom=transform[3] + (dataset.RasterYSize * projection.ystep),
        )

        super().__init__(
            area,
            projection,
            name=name
        )

        # The constructor works out the window from the area
        # so sanity check that the calculated window matches the
        # dataset's dimensions
        assert self.window == Window(0, 0, dataset.RasterXSize, dataset.RasterYSize)

        self._dataset = dataset
        self._dataset_path = Path(dataset.GetDescription())
        self._band = band
        self._raster_xsize = dataset.RasterXSize
        self._raster_ysize = dataset.RasterYSize
        self._ignore_nodata = ignore_nodata

    @property
    def _raster_dimensions(self) -> Tuple[int,int]:
        return (self._raster_xsize, self._raster_ysize)

    def close(self):
        try:
            if self._dataset:
                try:
                    self._dataset.Close()
                except AttributeError:
                    pass
                del self._dataset
        except AttributeError:
            # Don't error if close was already called
            pass

    def __getstate__(self) -> object:
        # Only support pickling on file backed layers (ideally read only ones...)
        if not self._dataset_path.exists():
            raise ValueError("Can not pickle layer that is not file backed.")
        odict = self.__dict__.copy()
        self._park()
        del odict['_dataset']
        return odict

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._unpark()

    def _park(self):
        if self._dataset is not None:
            self._dataset.Close()
        self._dataset = None

    def _unpark(self):
        if getattr(self, "_dataset", None) is None:
            try:
                self._dataset = gdal.Open(self._dataset_path)
            except RuntimeError as exc:
                raise FileNotFoundError(f"Failed to open pickled raster {self._dataset_path}") from exc

    @property
    def datatype(self) -> DataType:
        if self._dataset is None:
            self._unpark()
        assert self._dataset
        return DataType.of_gdal(self._dataset.GetRasterBand(1).DataType)

    @property
    def nodata(self) -> Optional[Any]:
        if self._dataset is None:
            self._unpark()
        assert self._dataset
        return self._dataset.GetRasterBand(self._band).GetNoDataValue()

    def _read_array_with_window(
        self,
        xoffset: int,
        yoffset: int,
        xsize: int,
        ysize: int,
        window: Window,
    ) -> Any:
        if self._dataset is None:
            self._unpark()
        assert self._dataset

        if (xsize <= 0) or (ysize <= 0):
            raise ValueError("Request dimensions must be positive and non-zero")

        # if we're dealing with an intersection, we can just read the data directly,
        # otherwise we need to read the data into another array with suitable padding
        target_window = Window(
            window.xoff + xoffset,
            window.yoff + yoffset,
            xsize,
            ysize
        )
        source_window = Window(
            xoff=0,
            yoff=0,
            xsize=self._raster_xsize,
            ysize=self._raster_ysize,
        )
        try:
            intersection = Window.find_intersection([source_window, target_window])
        except ValueError:
            return backend.zeros((ysize, xsize))

        if target_window == intersection:
            # The target window is a subset of or equal to the source, so we can just ask for the data
            data = backend.promote(self._dataset.GetRasterBand(self._band).ReadAsArray(*intersection.as_array_args))
        else:
            # We should read the intersection from the array, and the rest should be zeros
            subset = backend.promote(self._dataset.GetRasterBand(self._band).ReadAsArray(*intersection.as_array_args))
            region = np.array((
                (
                    (intersection.yoff - window.yoff) - yoffset,
                    (ysize - ((intersection.yoff - window.yoff) + intersection.ysize)) + yoffset,
                ),
                (
                    (intersection.xoff - window.xoff) - xoffset,
                    xsize - ((intersection.xoff - window.xoff) + intersection.xsize) + xoffset,
                )
            )).astype(int)
            data = backend.pad(subset, region, mode='constant')

        nodata = self.nodata
        if not self._ignore_nodata and nodata is not None:
            data = backend.where(data == nodata, float("nan"), data)

        return data



import os
import tempfile

import numpy as np
import pytest
import torch

import yirgacheffe
from tests.helpers import gdal_dataset_with_data
from yirgacheffe.layers import RasterLayer
from yirgacheffe.operators import LayerOperation

# These tests are marked skip for MLX, because there seems to be a problem with
# calling mx.eval in the tests for parallel save on Linux (which is what we use
# for github actions for instance). They do pass on macOS - but that could also
# just be down to Python versions. To add further complication, if I just run
# this file along on linux with MLX it passes fine 🤦
#
# It seems that under the hood MLX is doing some threading of its own and my
# guess is that that's interacting with the Python threading here.

def test_add_byte_layers_with_one_thread_uses_regular_save(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 4)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()

            path2 = os.path.join(tempdir, "test2.tif")
            data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80]])
            dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
            dataset2.Close()

            layer1 = RasterLayer.layer_from_file(path1)
            layer2 = RasterLayer.layer_from_file(path2)
            result = RasterLayer.empty_raster_layer_like(layer1)

            comp = layer1 + layer2
            with pytest.raises(TypeError):
                comp.parallel_save(result)


@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_add_byte_layers(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()

            path2 = os.path.join(tempdir, "test2.tif")
            data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80]])
            dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
            dataset2.Close()

            layer1 = RasterLayer.layer_from_file(path1)
            layer2 = RasterLayer.layer_from_file(path2)
            result = RasterLayer.empty_raster_layer_like(layer1)

            comp = layer1 + layer2
            comp.parallel_save(result)

            expected = data1 + data2
            actual = result.read_array(0, 0, 4, 2)

            assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_add_byte_layers_and_sum(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()

            path2 = os.path.join(tempdir, "test2.tif")
            data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80]])
            dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
            dataset2.Close()

            layer1 = RasterLayer.layer_from_file(path1)
            layer2 = RasterLayer.layer_from_file(path2)
            result = RasterLayer.empty_raster_layer_like(layer1)

            comp = layer1 + layer2
            sum_total = comp.parallel_save(result, and_sum=True)

            expected = data1 + data2
            actual = result.read_array(0, 0, 4, 2)

            assert (expected == actual).all()
            assert sum_total == expected.sum()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_sum(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()

            path2 = os.path.join(tempdir, "test2.tif")
            data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80]])
            dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
            dataset2.Close()

            layer1 = RasterLayer.layer_from_file(path1)
            layer2 = RasterLayer.layer_from_file(path2)

            comp = layer1 + layer2
            sum_total = comp.parallel_sum()

            expected = data1 + data2
            assert sum_total == expected.sum()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
@pytest.mark.parametrize("skip,expected_steps", [
    (1, [0.0, 0.25, 0.5, 0.75, 1.0]),
    (2, [0.0, 0.5, 1.0]),
])
def test_parallel_with_different_skip(monkeypatch, skip, expected_steps) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [1, 2, 3, 4], [5, 6, 7, 8]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()

            path2 = os.path.join(tempdir, "test2.tif")
            data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [10, 20, 30, 40], [50, 60, 70, 80]])
            dataset2 = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=path2)
            dataset2.Close()

            layer1 = RasterLayer.layer_from_file(path1)
            layer2 = RasterLayer.layer_from_file(path2)
            result = RasterLayer.empty_raster_layer_like(layer1)

            callback_possitions = []

            comp = layer1 + layer2
            comp.ystep = skip
            comp.parallel_save(result, callback=callback_possitions.append)

            expected = data1 + data2
            actual = result.read_array(0, 0, 4, 4)

            assert (expected == actual).all()

            assert callback_possitions == expected_steps

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_equality(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[1, 2, 3, 4], [4, 3, 2, 1]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()
            with RasterLayer.layer_from_file(path1) as layer1:
                with RasterLayer.empty_raster_layer_like(layer1) as result:
                    comp = layer1 == 2
                    comp.parallel_save(result)

                    expected = np.array([[0, 1, 0, 0], [0, 0, 1, 0]])
                    actual = result.read_array(0, 0, 4, 2)

                    assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_equality_to_file(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            path2 = os.path.join(tempdir, "result.tif")
            data1 = np.array([[1, 2, 3, 4], [4, 3, 2, 1]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()
            with RasterLayer.layer_from_file(path1) as layer1:
                comp = layer1 == 2
                with RasterLayer.empty_raster_layer_like(layer1, filename=path2) as result:
                    comp.parallel_save(result)
            with RasterLayer.layer_from_file(path2) as actual_result:
                expected = np.array([[0, 1, 0, 0], [0, 0, 1, 0]])
                actual = actual_result.read_array(0, 0, 4, 2)
                assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_unary_numpy_apply_with_function(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()
            layer1 = RasterLayer.layer_from_file(path1)

            result = RasterLayer.empty_raster_layer_like(layer1)

            def simple_add(chunk):
                return chunk + 1.0

            comp = layer1.numpy_apply(simple_add)
            comp.ystep = 1
            comp.parallel_save(result)

            expected = data1 + 1.0
            actual = result.read_array(0, 0, 4, 2)

            assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_unary_numpy_apply_with_lambda(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()
            layer1 = RasterLayer.layer_from_file(path1)

            result = RasterLayer.empty_raster_layer_like(layer1)

            comp = layer1.numpy_apply(lambda a: a + 1.0)
            comp.ystep = 1
            comp.parallel_save(result)

            expected = data1 + 1.0
            actual = result.read_array(0, 0, 4, 2)

            assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_where_simple(monkeypatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(yirgacheffe.constants, "YSTEP", 1)
        m.setattr(LayerOperation, "save", None)
        with tempfile.TemporaryDirectory() as tempdir:
            path1 = os.path.join(tempdir, "test1.tif")
            data1 = np.array([[0, 1, 0, 2], [0, 0, 1, 1]])
            dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
            dataset1.Close()
            layer1 = RasterLayer.layer_from_file(path1)

            result = RasterLayer.empty_raster_layer_like(layer1)

            comp = LayerOperation.where(layer1 > 0, 1, 2)
            comp.ystep = 1
            comp.parallel_save(result)

            expected = np.where(data1 > 0, 1, 2)
            actual = result.read_array(0, 0, 4, 2)
            assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_conv2d() -> None:
    with tempfile.TemporaryDirectory() as tempdir:

        data1 = np.array([
            [0, 0, 0, 0, 0],
            [0, 1, 1, 1, 0],
            [0, 1, 1, 1, 0],
            [0, 1, 1, 1, 0],
            [0, 0, 0, 0, 0],
        ]).astype(np.float32)
        weights = np.array([
            [0.0, 0.1, 0.0],
            [0.1, 0.6, 0.1],
            [0.0, 0.1, 0.0],
        ])

        conv = torch.nn.Conv2d(1, 1, 3, padding=1, bias=False)
        conv.weight = torch.nn.Parameter(torch.from_numpy(np.array([[weights.astype(np.float32)]])))
        tensorres = conv(torch.from_numpy(np.array([[data1]])))
        expected = tensorres.detach().numpy()[0][0]

        path1 = os.path.join(tempdir, "test1.tif")
        dataset1 = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=path1)
        dataset1.Close()

        with RasterLayer.layer_from_file(path1) as layer1:

            calc = layer1.conv2d(weights)
            with RasterLayer.empty_raster_layer_like(layer1) as res:
                calc.save(res)
                actual = res.read_array(0, 0, 5, 5)

                # Torch and MLX give slightly different rounding
                assert np.isclose(expected, actual).all()



import os
import tempfile

import numpy as np
import pytest

import yirgacheffe
from tests.helpers import gdal_dataset_with_data, make_vectors_with_mutlile_ids
from yirgacheffe.layers import ConstantLayer, RasterLayer, VectorLayer
from yirgacheffe.window import Area

def test_add_windows() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data2))

    assert layer1.area != layer2.area
    assert layer1.window != layer2.window

    calc = layer1 + layer2

    assert calc.area == layer2.area
    assert calc.window == layer2.window

    result = RasterLayer.empty_raster_layer_like(calc)
    calc.save(result)

    expected = np.array([[11, 22, 30, 40], [53, 64, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])
    actual = result.read_array(0, 0, 4, 4)
    assert (expected == actual).all()

def test_multiply_windows() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data2))

    assert layer1.area != layer2.area
    assert layer1.window != layer2.window

    calc = layer1 * layer2

    assert calc.area == layer1.area
    assert calc.window == layer1.window

    result = RasterLayer.empty_raster_layer_like(calc)
    calc.save(result)

    expected = np.array([[10, 40], [150, 240]])
    actual = result.read_array(0, 0, 2, 2)
    assert (expected == actual).all()

def test_add_windows_offset() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

    layer1 = RasterLayer(gdal_dataset_with_data((-0.02, 0.02), 0.02, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((-0.04, 0.04), 0.02, data2))

    assert layer1.area != layer2.area
    assert layer1.window != layer2.window

    calc = layer1 + layer2

    assert calc.area == layer2.area

    result = RasterLayer.empty_raster_layer_like(calc)
    calc.save(result)

    expected = np.array([[10, 20, 30, 40], [50, 61, 72, 80], [90, 103, 114, 120], [130, 140, 150, 160]])
    actual = result.read_array(0, 0, 4, 4)
    assert (expected == actual).all()

def test_multiply_windows_offset() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

    layer1 = RasterLayer(gdal_dataset_with_data((-0.02, 0.02), 0.02, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((-0.04, 0.04), 0.02, data2))

    assert layer1.area != layer2.area
    assert layer1.window != layer2.window

    calc = layer1 * layer2

    assert calc.area == layer1.area

    result = RasterLayer.empty_raster_layer_like(calc)
    calc.save(result)

    expected = np.array([[60, 140], [300, 440]])
    actual = result.read_array(0, 0, 2, 2)
    assert (expected == actual).all()

def test_add_windows_sum() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

    layer1 = RasterLayer(gdal_dataset_with_data((-0.02, 0.02), 0.02, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((-0.04, 0.04), 0.02, data2))

    calc = layer1 + layer2
    total = calc.sum()

    expected = np.array([[10, 20, 30, 40], [50, 61, 72, 80], [90, 103, 114, 120], [130, 140, 150, 160]])
    assert total == np.sum(expected)

def test_multiply_windows_sum() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

    layer1 = RasterLayer(gdal_dataset_with_data((-0.02, 0.02), 0.02, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((-0.04, 0.04), 0.02, data2))

    calc = layer1 * layer2
    total = calc.sum()

    expected = np.array([[60, 140], [300, 440]])
    assert total == np.sum(expected)

def test_constant_layer_result_rhs_add() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
    const_layer = ConstantLayer(1.0)

    calc = layer1 + const_layer

    assert calc.area == layer1.area

    result = RasterLayer.empty_raster_layer_like(calc)
    calc.save(result)
    actual = result.read_array(0, 0, 4, 2)

    expected = 1.0 + data1

    assert (expected == actual).all()

def test_constant_layer_result_lhs_add() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
    const_layer = ConstantLayer(1.0)
    result = RasterLayer.empty_raster_layer_like(layer1)

    intersection = RasterLayer.find_intersection([layer1, const_layer])
    const_layer.set_window_for_intersection(intersection)
    layer1.set_window_for_intersection(intersection)

    calc = const_layer + layer1

    assert calc.area == layer1.area

    result = RasterLayer.empty_raster_layer_like(calc)
    calc.save(result)
    actual = result.read_array(0, 0, 4, 2)

    expected = 1.0 + data1

    assert (expected == actual).all()

def test_constant_layer_result_rhs_multiply() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
    const_layer = ConstantLayer(2.0)

    calc = layer1 * const_layer

    assert calc.area == layer1.area

    result = RasterLayer.empty_raster_layer_like(calc)
    calc.save(result)
    actual = result.read_array(0, 0, 4, 2)

    expected = data1 * 2.0

    assert (expected == actual).all()

def test_constant_layer_result_lhs_multiply() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]])
    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
    const_layer = ConstantLayer(2.0)
    result = RasterLayer.empty_raster_layer_like(layer1)

    intersection = RasterLayer.find_intersection([layer1, const_layer])
    const_layer.set_window_for_intersection(intersection)
    layer1.set_window_for_intersection(intersection)

    calc = const_layer * layer1

    assert calc.area == layer1.area

    result = RasterLayer.empty_raster_layer_like(calc)
    calc.save(result)
    actual = result.read_array(0, 0, 4, 2)

    expected = 2.0 * data1

    assert (expected == actual).all()

def test_vector_layers_add() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 1.0, data1))

    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.gpkg")
        areas = {
            (Area(-10.0, 10.0, 0.0, 0.0), 42),
            (Area(0.0, 0.0, 10, -10), 43)
        }
        make_vectors_with_mutlile_ids(areas, path)

        burn_value = 2
        layer2 = VectorLayer.layer_from_file(path, None, layer1.pixel_scale, layer1.projection, burn_value=burn_value)
        layer2_total = layer2.sum()
        assert layer2_total == ((layer2.window.xsize * layer2.window.ysize) / 2) * burn_value

        calc = layer1 + layer2

        assert calc.area == layer2.area

        total = calc.sum()
        assert total == layer2_total + np.sum(data1)

def test_vector_layers_multiply() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    layer1 = RasterLayer(gdal_dataset_with_data((-1.0, 1.0), 1.0, data1))

    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, "test.gpkg")
        areas = {
            (Area(-10.0, 10.0, 0.0, 0.0), 42),
            (Area(0.0, 0.0, 10, -10), 43)
        }
        make_vectors_with_mutlile_ids(areas, path)

        burn_value = 2
        layer2 = VectorLayer.layer_from_file(path, None, layer1.pixel_scale, layer1.projection, burn_value=burn_value)
        layer2_total = layer2.sum()
        assert layer2_total == ((layer2.window.xsize * layer2.window.ysize) / 2) * burn_value

        calc = layer1 * layer2

        assert calc.area == layer1.area

        result = RasterLayer.empty_raster_layer_like(calc)
        calc.save(result)
        actual = result.read_array(0, 0, 2, 2)

        expected = np.array([[2, 0], [0, 8]])
        assert (expected == actual).all()

@pytest.mark.skipif(yirgacheffe._backends.BACKEND != "NUMPY", reason="Only applies for numpy")
def test_parallel_save_windows() -> None:
    data1 = np.array([[1, 2], [3, 4]])
    data2 = np.array([[10, 20, 30, 40], [50, 60, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])

    with tempfile.TemporaryDirectory() as tempdir:
        layer1_filename = os.path.join(tempdir, "layer1.tif")
        layer1_dataset = gdal_dataset_with_data((0.0, 0.0), 0.02, data1, filename=layer1_filename)
        layer1_dataset.Close()

        layer2_filename = os.path.join(tempdir, "layer2.tif")
        layer2_dataset = gdal_dataset_with_data((0.0, 0.0), 0.02, data2, filename=layer2_filename)
        layer2_dataset.Close()

        layer1 = RasterLayer.layer_from_file(layer1_filename)
        layer2 = RasterLayer.layer_from_file(layer2_filename)

        assert layer1.area != layer2.area
        assert layer1.window != layer2.window

        calc = layer1 + layer2

        assert calc.area == layer2.area
        assert calc.window == layer2.window

        result = RasterLayer.empty_raster_layer_like(calc)
        calc.parallel_save(result)

        expected = np.array([[11, 22, 30, 40], [53, 64, 70, 80], [90, 100, 110, 120], [130, 140, 150, 160]])
        actual = result.read_array(0, 0, 4, 4)
        assert (expected == actual).all()



import pytest
from osgeo import gdal

from tests.helpers import gdal_dataset_of_region, gdal_empty_dataset_of_region
from yirgacheffe.window import Area, MapProjection, Window
from yirgacheffe.layers import RasterLayer, ConstantLayer, H3CellLayer
from yirgacheffe import WGS_84_PROJECTION


def test_find_intersection_empty_list() -> None:
    with pytest.raises(ValueError):
        _ = RasterLayer.find_intersection([])

def test_find_intersection_single_item() -> None:
    layer = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    intersection = RasterLayer.find_intersection([layer])
    assert intersection == layer.area

def test_find_intersection_same() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    ]
    intersection = RasterLayer.find_intersection(layers)
    assert intersection == layers[0].area

def test_find_intersection_subset() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(-1, 1, 1, -1), 0.02))
    ]
    intersection = RasterLayer.find_intersection(layers)
    assert intersection == layers[1].area

def test_find_intersection_overlap() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(-15, 15, -5, -5), 0.02))
    ]
    intersection = RasterLayer.find_intersection(layers)
    assert intersection == Area(-10, 10, -5, -5)

def test_find_intersection_distinct() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-110, 10, -100, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(100, 10, 110, -10), 0.02))
    ]
    with pytest.raises(ValueError):
        _ = RasterLayer.find_intersection(layers)

def test_find_intersection_with_constant() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        ConstantLayer(1.0)
    ]
    intersection = RasterLayer.find_intersection(layers)
    assert intersection == layers[0].area

def test_find_intersection_different_pixel_pitch() -> None:
    layers = [
        RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02)),
        RasterLayer(gdal_dataset_of_region(Area(-15, 15, -5, -5), 0.01))
    ]
    with pytest.raises(ValueError):
        _ = RasterLayer.find_intersection(layers)

@pytest.mark.parametrize("scale", [0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09])
def test_set_intersection_self(scale) -> None:
    layer = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), scale))
    old_window = layer.window

    # note that the area we passed to gdal_dataset_of_region isn't pixel aligned, so we must
    # use the area from loading the dataset
    layer.set_window_for_intersection(layer.area)
    assert layer.window == old_window

    # reset should not do much here
    layer.reset_window()
    assert layer.window == old_window

def test_set_intersection_subset() -> None:
    layer = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    assert layer.window == Window(0, 0, 1000, 1000)
    origin_before_pixel = layer.read_array(0, 0, 1, 1)

    intersection = Area(-1.0, 1.0, 1.0, -1.0)

    layer.set_window_for_intersection(intersection)
    assert layer.window == Window(450, 450, 100, 100)
    origin_after_pixel = layer.read_array(0, 0, 1, 1)

    # The default data is populated as a mod of row value, so given
    # were not a multiple of 256 off, these pixels should not have the same
    # value in them
    assert origin_before_pixel[0][0] != origin_after_pixel[0][0]

    layer.reset_window()
    assert layer.window == Window(0, 0, 1000, 1000)

def test_set_intersection_distinct() -> None:
    layer = RasterLayer(gdal_dataset_of_region(Area(-10, 10, 10, -10), 0.02))
    intersection = Area(-101.0, 1.0, -100.0, -1.0)
    with pytest.raises(ValueError):
        layer.set_window_for_intersection(intersection)

def test_find_intersection_nearly_same() -> None:
    # This testcase is based on a real instance we hit whereby
    # the layers were effectively the same, and intended to be the same,
    # but a rounding error of less than the floating point epsilon was in
    # one of the files.
    #
    # gdalinfo rounds the numbers, so it wasn't obvious, but inspecting the
    # GEOTiffs with tifffile (a pure tiff library) showed the error, and GDAL
    # in python showed the error too.
    #
    # Yirgacheffe at the time blew up as it knew to ignore the difference
    # when doing a comparison (thanks to layers.py::almost_equal(a,b)), but
    # when you then multiplied this up by the area it rounded poorly.
    layers = [
        RasterLayer(gdal_empty_dataset_of_region(
            Area(left=-180.00082337073326, top=90.00041168536663, right=180.00082337073326, bottom=-90.00041168536663),
            0.0008983152841195215
        )),
        RasterLayer(gdal_empty_dataset_of_region(
            Area(left=-180.00082337073326, top=90.00041168536661, right=180.00082337073326, bottom=-90.00041168536664),
            0.0008983152841195216
        )),
        RasterLayer(gdal_empty_dataset_of_region(
            Area(left=-180, top=90.00041168536661, right=180, bottom=-90.00041168536664),
            0.0008983152841195215
        )),
        RasterLayer(gdal_empty_dataset_of_region(
            Area(left=-3.6372785853999425, top=47.767016917771436, right=3.578888091932174, bottom=42.068104755317194),
            0.0008983152841195215
        ))
    ]

    intersection = RasterLayer.find_intersection(layers)
    assert intersection == layers[-1].area
    for layer in layers:
        layer.set_window_for_intersection(intersection)
    for other in layers[1:]:
        assert layers[0].window.xsize == other.window.xsize
        assert layers[0].window.ysize == other.window.ysize

def test_intersection_stability():
    # This test uses h3 tiles as a lazy way to get some bounded regions,
    # but the bug this test exercises is not h3 specific. This was another case of
    # a rounding error that causes set_window_for_* methods to wobble depending on how far
    # away from the top left thing where. adding round_down_pixels fixed this.
    cells = ["874b93aaeffffff", "874b93a85ffffff", "874b93aa3ffffff", "874b93a84ffffff", "874b93a80ffffff"]
    projection = MapProjection(WGS_84_PROJECTION, 0.000898315284120,-0.000898315284120)

    tiles = [
        H3CellLayer(cell_id, projection)
    for cell_id in cells]

    # composing the same tiles within different areas should not cause them to
    # wobble around
    union = RasterLayer.find_union(tiles)
    superunion = union.grow(0.02)

    scratch1 = RasterLayer.empty_raster_layer(union, projection.scale, gdal.GDT_Float64, name='s1')
    scratch2 = RasterLayer.empty_raster_layer(superunion, projection.scale, gdal.GDT_Float64, name='s2')

    relative_offsets = {}

    for scratch in [scratch1, scratch2]:
        offsets = []
        first = None
        for tile in tiles:
            scratch.reset_window()
            layers = [scratch, tile]
            intersection = RasterLayer.find_intersection(layers)

            # We know the tile is a subset of the scratch region, so
            # the intersection should just be that
            assert intersection == tile.area

            for layer in layers:
                layer.set_window_for_intersection(intersection)

            if first is None:
                first = scratch.window
            else:
                offset = scratch.window.xoff - first.xoff, scratch.window.yoff - first.yoff
                offsets.append(offset)
        relative_offsets[scratch.name] = offsets

    assert relative_offsets[scratch1.name] == relative_offsets[scratch2.name]




import numpy as np
import pytest
from osgeo import gdal

from yirgacheffe.operators import DataType
from yirgacheffe._backends import backend, BACKEND
from yirgacheffe.layers import RasterLayer

from tests.helpers import gdal_dataset_with_data

@pytest.mark.parametrize("gtype", [
    gdal.GDT_Int8,
    gdal.GDT_Int16,
    gdal.GDT_Int32,
    gdal.GDT_Int64,
    gdal.GDT_Byte,
    gdal.GDT_UInt16,
    gdal.GDT_UInt32,
    gdal.GDT_UInt64,
    gdal.GDT_Float32,
])
def test_round_trip(gtype) -> None:
    ytype = DataType.of_gdal(gtype)
    backend_type = backend.dtype_to_backed(ytype)
    assert backend.backend_to_dtype(backend_type) == ytype

@pytest.mark.parametrize("ytype", [
    DataType.Int8,
    DataType.Int16,
    DataType.Int32,
    DataType.Int64,
    DataType.UInt8,
    DataType.UInt16,
    DataType.UInt32,
    DataType.UInt64,
    DataType.Float32,
    DataType.Float64,
])
def test_round_trip_from_gdal(ytype) -> None:
    gtype = ytype.to_gdal()
    assert DataType.of_gdal(gtype) == ytype

def test_round_trip_float64() -> None:
    backend_type = backend.dtype_to_backed(DataType.Float64)
    ytype = backend.backend_to_dtype(backend_type)
    match BACKEND:
        case "NUMPY":
            assert ytype == DataType.Float64
        case "MLX":
            assert ytype == DataType.Float32

def test_float_to_int() -> None:
    data1 = np.array([[1.0, 2.0, 3.0, 4.0], [5.5, 6.5, 7.5, 8.5]])

    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 0.02, data1))
    # Note the float 32 here is to rule out that writing the result to the
    # new dataset was what caused the truncation
    result = RasterLayer.empty_raster_layer_like(layer1, datatype=DataType.Float32)

    comp = layer1.astype(DataType.UInt8)
    comp.save(result)

    expected = backend.promote(np.array([[1, 2, 3, 4], [5, 6, 7, 8]]))
    actual = backend.demote_array(result.read_array(0, 0, 4, 2))
    assert (expected == actual).all()
